static const char rcsid[] = "$Id: crlfmaildir.c,v 1.11 2000/09/30 19:52:35 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * Maildir format mailboxes, EXCEPT:
 *
 * messages have lines ending in \r\n so we don't need to mangle them
 * messages are already bytes stuffed.
 * messages have ":<size without bytestuffing>" at the end of their names.
 *
 * we don't bother moving stuff from "new", and ignore "cur"
 * (infact "cur" dosn't exist)
 *
 * jasper@ivision.co.uk//jasper@pointless.net
 *
 */


#include "includes.h"

#include "maildrop.h"
#include "crlfmaildir.h"
#include "const.h"
#include "md5.h"
#include "log.h"
#include "fdfgets.h"
#include "cmds.h"
#include "sktbuff.h"

#include <errno.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) (strlen((dirent)->d_name))
#else
# define dirent direct
# define NAMLEN(dirent) ((dirent)->d_namlen)
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

/* from gettext package */
#if (defined HAVE_MMAP && defined HAVE_MUNMAP && !defined DISALLOW_MMAP) \
    || (defined _LIBC && defined _POSIX_MAPPED_FILES)
# include <sys/mman.h>
# undef HAVE_MMAP
# define HAVE_MMAP      1
#else
# undef HAVE_MMAP
# error "you must have mmap() support on your system to use crlfmaildir"
#endif

#ifdef SPENCRYPTION
extern int enctrans, encenabled, encpipe[];
#endif
	
extern void check_wccount(void);

extern char maildrop_name[];
extern char username[];
extern int msgnr, maxmsgnr, msgdel;
extern struct message *messages;
int msgfd;
int crlf_mdir_exists;
#ifdef BULLETINS
static pid_t nomsg = 0, nnomsg = 0;
#endif
#ifdef CREATEMAILDROP
extern int createmaildrop;
#endif

struct str_maildrop crlf_mdir_maildrop = {
	crlf_mdir_init,
	crlf_mdir_update,
	crlf_mdir_release,
	crlf_mdir_retrieve,
	crlf_mdir_top,
#ifndef BULLETINS
	crlf_mdir_md5_uidl_message
#else
	crlf_mdir_md5_uidl_message,
	crlf_mdir_add_message,
	crlf_mdir_end_of_adding
#endif
};

int crlf_mdir_compare(const void *arg1, const void *arg2) {
        time_t d1, d2;
    
	d1 = ((const struct message *)(arg1))->msg_time; 
        d2 = ((const struct message *)(arg2))->msg_time;
	if (d1 < d2)
	        return -1;
        if (d1 == d2)
		return 0;
	return 1;
}

#ifdef BULLETINS
void crlf_mdir_end_of_adding(void) {
	if (crlf_mdir_exists == 0)
		return;
	qsort(messages, msgnr, sizeof(struct message), crlf_mdir_compare);
}

int crlf_mdir_add_message(int fd) {
	char filename[PATH_MAX];
	char nfilename[PATH_MAX];
	char msgname[100 + 256];
	char hstname[256];
	pid_t tmppid = getpid() + nomsg;
	pid_t ntmppid = getpid() + nnomsg;
	time_t tmptime = time(NULL);
	int nmsgfd, nl = 1;
	ssize_t mcount;
	char *linebegin;
	char linebuf[128 + 2]; /* first char - optional '.'
				  last chars - optional \r\n */
	linebuf[0] = '.';
	if (crlf_mdir_exists == 0) {
		pop_log(pop_priority, "crlfmaildir: can't add bulletin "
			"- maildir doesn't exist");
		return 0;
	}
	if (gethostname(hstname, sizeof(hstname)) < 0) {
		close(fd);
		pop_log(pop_priority, "crlfmaildir: can't get name of host");
		pop_error("crlfmaildir: gethostname");
		send_error("fatal error");
		exit(1);
	}
	do {
		snprintf(msgname, sizeof(msgname), "tmp/%lu.%lu.%.256s",
			(unsigned long int)tmptime, (unsigned long int)tmppid,
			hstname);
		if ((strlen(msgname) + strlen(maildrop_name) + 2) >
			sizeof(filename)) {
			close(fd);
			pop_log(pop_priority, "crlfmaildir: message file name "
				"is too long");
			send_error("fatal error");
			exit(1);
		}
		strcpy(filename, maildrop_name);
		strcat(filename, "/");
		strcat(filename, msgname);
		nmsgfd = open(filename, O_WRONLY | O_CREAT | O_EXCL, 0600);
		if ((nmsgfd < 0) && (errno == EEXIST)) {
			tmppid++;
			nomsg++;
			if (nomsg > 500) {
				close(fd);
				pop_log(pop_priority, "crlfmaildir: can't find "
					"unused message file name");
				send_error("fatal error");
				exit(1);
			}
			continue; /* try next message file name */
		}
		if (nmsgfd < 0) {
			close(fd);
			pop_log(pop_priority, "crlfmaildir: can't create "
				"message file: %.1024s", filename);
			pop_error("crlfmaildir: open");
			send_error("fatal error");
			exit(1);			
		}
		nomsg++;
		break;
	} while(1);
	fd_initfgets();
	if ((mcount = fd_fgets(linebuf + 1, sizeof(linebuf) - 2, fd,
		NULL)) < 0) {
		close(nmsgfd);
		close(fd);
		unlink(filename);
		pop_log(pop_priority, "crlfmaildir: can't read message");
		pop_error("crlfmaildir: read");
		send_error("fatal error");
		exit(1);	
	}
	if (mcount > 5)
		if (strncmp(linebuf + 1, "From ", 5) == 0) {
			while (linebuf[mcount] != '\n') {
				if ((mcount = fd_fgets(linebuf + 1,
					sizeof(linebuf) - 2, fd, NULL)) < 0) {
					close(nmsgfd);
					close(fd);
					unlink(filename);
					pop_log(pop_priority, "crlfmaildir: "
						"can't read message");
					pop_error("crlfmaildir: read");
					send_error("fatal error");
					exit(1);
				}
				if (mcount == 0)
					break;
			}
			if (mcount == 0) { /* null message */
				close(nmsgfd);
				close(fd);
				unlink(filename);
				pop_log(pop_priority, "crlfmaildir: empty "
					"message");
				send_error("fatal error");
				exit(1);
			}
			mcount = fd_fgets(linebuf + 1, sizeof(linebuf) - 2,
				fd, NULL);
		}
	if (md_realloc(sizeof(struct crlf_mdir_message)) < 0) {
		close(nmsgfd);
		close(fd);
		unlink(filename);
		pop_log(pop_priority, "crlfmaildir: no memory available");
		send_error("no memory available");
		exit(1);
	}
	msgnr++;
	messages[msgnr - 1].msg_time = tmptime;
	messages[msgnr - 1].deleted = messages[msgnr - 1].crlfsize =
		messages[msgnr - 1].size = messages[msgnr - 1].cread = 0;
	while (mcount > 0) {
		if (linebuf[mcount] == '\n') {
			linebuf[mcount] = '\r';
			linebuf[mcount + 1] = '\n'; 
			/* no overflow - mcount <= sizeof(linebuf) - 2 */
			mcount++;
		}
		messages[msgnr - 1].crlfsize += mcount;
		messages[msgnr - 1].size += mcount;
		if ((linebuf[1] == '.') && (nl)) {
			messages[msgnr - 1].size++;
			mcount++;
			linebegin = linebuf; /* linebuf[0] = '.' */
		} else
			linebegin = linebuf + 1;
		if (write(nmsgfd, linebegin, mcount) != mcount) {
			close(nmsgfd);
			close(fd);
			unlink(filename);
			pop_log(pop_priority, "crlfmaildir: can't write to "
				"message file: %.1024s", filename);
			pop_error("crlfmaildir: write");
			send_error("fatal errror");
			exit(1);
		}
		nl = (linebegin[mcount - 1] == '\n') ? 1 : 0;
		mcount = fd_fgets(linebuf + 1, sizeof(linebuf) - 2, fd, NULL);
	}
	if (mcount < 0) {
		close(nmsgfd);
		close(fd);
		unlink(filename);
		pop_log(pop_priority, "crlfmaildir: can't read from message "
			"file: %.1024s", filename);
		pop_error("crlfmaildir: read");
		send_error("can't read from message file");
		exit(1);
	}
	if (!nl) {
		close(nmsgfd);
		close(fd);
		unlink(filename);
		pop_log(pop_priority, "crlfmaildir: message is damaged: "
			"%.1024s", filename);
		send_error("message is damaged");
		exit(1);
	}
	if (close(nmsgfd) < 0) {
		close(fd);
		unlink(filename);
		pop_log(pop_priority, "crlfmaildir: can't close message file: "
			"%.1024s", filename);
		pop_error("crlfmaildir: close");
		send_error("can't close message file");
		exit(1);
	}
	do {
		snprintf(msgname, sizeof(msgname), "new/%lu.%lu.%.256s:%lu",
			(unsigned long int)tmptime, (unsigned long int)ntmppid,
			hstname, (unsigned long int)messages[msgnr - 1].crlfsize);
		if ((strlen(msgname) + strlen(maildrop_name) + 2) >
			sizeof(nfilename)) {
			close(fd);
			unlink(filename);
			pop_log(pop_priority, "crlfmaildir: message file name "
				"is too long");
			send_error("fatal error");
			exit(1);
		}
		strcpy(nfilename, maildrop_name);
		strcat(nfilename, "/");
		strcat(nfilename, msgname);
		nmsgfd = link(filename, nfilename);
		if ((nmsgfd < 0) && (errno == EEXIST)) {
			ntmppid++;
			nnomsg++;
			if (nnomsg > 500) {
				close(fd);
				unlink(filename);
				pop_log(pop_priority, "crlfmaildir: can't find "
					"unused message file name");
				send_error("fatal error");
				exit(1);
			}
			continue; /* try next message file name */
		}
		if (nmsgfd < 0) {
			close(fd);
			unlink(filename);
			pop_log(pop_priority, "crlfmaildir: can't link message "
				"file: %.1024s", nfilename);
			pop_error("crlfmaildir: link");
			send_error("fatal error");
			exit(1);			
		}
		nnomsg++;
		break;
	} while(1);
	unlink(filename);
	if ((((struct crlf_mdir_message *)
		(messages[msgnr - 1].md_specific))->filename =
		(char *) malloc(strlen(msgname) + 1)) == NULL) {
		close(nmsgfd);
		close(fd);
		unlink(nfilename); /* remove message */
		pop_log(pop_priority, "crlfmaildir: no memory available");
		send_error("no memory available");
		exit(1);
	}
	strcpy(((struct crlf_mdir_message *)
		(messages[msgnr - 1].md_specific))->filename, msgname);
	return 1;
}
#endif

void crlf_mdir_clean(void) {
	struct stat stat_buf;
	DIR *dir;
	struct dirent *d_entry;
	char filename[PATH_MAX], *tmp;
	time_t act_time;
	
	if ((strlen(maildrop_name) + 7) > sizeof(filename))
		return;
	strcpy(filename, maildrop_name);
	strcat(filename, "/tmp/");
	tmp = filename + strlen(filename);
	act_time = time(NULL);
	if ((dir = opendir(filename)) == NULL)
		return;
	while ((d_entry = readdir(dir)) != NULL) {
		if (d_entry->d_name[0] == '.')
			continue;
		*tmp = 0;
		if ((strlen(filename) + NAMLEN(d_entry) + 1) >
			sizeof(filename))
			break;
		strcpy(tmp, d_entry->d_name);
		if (stat(filename, &stat_buf) < 0) {
			continue;
		}
		if (act_time > (stat_buf.st_atime + (36*3600)))
			unlink(filename);
	}
	closedir(dir);
}

void crlf_mdir_append(const char *directory) {
	DIR *dirstream;
	struct dirent *direntry;
	char filename[PATH_MAX];
	char *tmp;
	char *fname, *sizep;
	struct stat stbuf;
	
	if ((strlen(directory) + strlen(maildrop_name) + 4) >
		sizeof(filename)) {
		crlf_mdir_release();
		pop_log(pop_priority, "crlfmaildir: maildir name too long");
		send_error("maildir name too long");
		exit(1);
	}
	strcpy(filename, maildrop_name); /* e.g. /var/mail/user */
	strcat(filename, "/");  /* /var/mail/user/ */
	fname = filename + strlen(filename);
	strcat(filename, directory);	 /* /var/mail/user/new */
	if ((dirstream = opendir(filename)) == NULL) {
		pop_log(pop_priority, "crlfmaildir: can't open "
			"maildir: %.1024s", filename);
		pop_error("crlfmaildir: opendir");
#ifdef CREATEMAILDROP
		if ((createmaildrop) || (errno != ENOENT)) {
#else
		if (errno != ENOENT) {
#endif
			crlf_mdir_release();
			send_error("can't open maildir");
			exit(1);
		}
		crlf_mdir_exists = 0;
		return;
	}
	strcat(filename, "/");
	tmp = filename + strlen(filename);

	while ((direntry = readdir(dirstream)) != NULL) {
		if (direntry->d_name[0] == '.')
			continue;
		if ((NAMLEN(direntry) + strlen(filename) + 1) >
			sizeof(filename)) {
			closedir(dirstream);
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: message file name "
				"too long");
			send_error("message file name too long");
			exit(1);
		}
		if ( (sizep = strrchr(direntry->d_name, ':')) == NULL )
		{
			closedir(dirstream);
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: message %.1024s "
				"does NOT contain a ':' character",
				direntry->d_name);
			send_error("maildir possible corrupt");
			exit(1);
		}
		sizep++; /* skip the ':' */
		strcpy(tmp, direntry->d_name);
		if (md_realloc(sizeof(struct crlf_mdir_message)) < 0) {
			closedir(dirstream);
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: no memory "
				"available");
			send_error("no memory available");
			exit(1);
		}
		if ((((struct crlf_mdir_message *)
			(messages[msgnr].md_specific))->filename =
			(char *) malloc(strlen(fname) + 1)) == NULL) {
			closedir(dirstream);
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: no memory "
				"available");
			send_error("no memory available");
			exit(1);
		}
		strcpy(((struct crlf_mdir_message *)
			(messages[msgnr].md_specific))->filename, fname);
		messages[msgnr].deleted = messages[msgnr].size =
			messages[msgnr].cread = 0;
			 /* this probably isn't needed, but dosn't hurt */
		messages[msgnr].crlfsize = atoi(sizep);
		/* calculated by into_crlfmaildir or crlf_mdir_add_message() */
		if (messages[msgnr].crlfsize == 0)
		{
			closedir(dirstream);
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: got 0 from "
				"atoi %.1024s :: %.256s", sizep,
				strerror(errno));
			send_error("maildir possible corrupt");
			exit(1);
		}
		if (stat(filename, &stbuf) < 0) {
			closedir(dirstream);
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: stat failed: "
				"%.1024s", filename);
			pop_error("crlfmaildir: stat");
			send_error("fatal error");
			exit(1);
		}
		messages[msgnr].msg_time = stbuf.st_mtime;
		messages[msgnr].size = stbuf.st_size;
		msgnr++;
		*tmp = 0;
	}
	closedir(dirstream);
}

#ifdef CREATEMAILDROP
void crlf_mdir_mkdir(char *dirname) {
	if (mkdir(dirname, 0700) < 0) {
		if (errno != EEXIST)
		{
			pop_log(pop_priority, "crlfmaildir: can't create "
				"directory: %.1024s", dirname);
			pop_error("crlfmaildir: mkdir");
			send_error("fatal error");
			exit(1);
		}
	}
}

void crlf_mdir_create(void) {
	char mdrop[PATH_MAX], newdir[PATH_MAX], *tmp;
	struct stat stbuf;
	int x, newdirlen;
	
	if ((strlen(maildrop_name) + 5) > sizeof(mdrop)) {
		crlf_mdir_release();
		pop_log(pop_priority, "crlfmaildir: maildrop name too long");
		send_error("fatal error");
		exit(1);
	}
	strcpy(mdrop, maildrop_name);
	tmp = mdrop + strlen(mdrop);

	if (stat(mdrop, &stbuf) == 0)
		return;

	strcpy(newdir, maildrop_name);
	newdirlen = strlen(newdir);
	for (x = newdirlen ; x > 0 ; x--)
	{
		if (newdir[x] == '/')
			newdir[x] = '\0';
		else
			continue;
		if (stat(newdir, &stbuf) == 0)
		{
			break; /* we've found the top most dir that exists */
		}
	}

	for (x = 0 ; x < newdirlen ; x++)
	{
		if (newdir[x] == '\0')
		{
			newdir[x] = '/';
			pop_log(pop_priority, "crlfmaildir: makeing %s",
				newdir);
			crlf_mdir_mkdir(newdir);
		}
	}

	strcpy(tmp, "/new");
	crlf_mdir_mkdir(mdrop);
	strcpy(tmp, "/cur");
	crlf_mdir_mkdir(mdrop);
	strcpy(tmp, "/tmp");
	crlf_mdir_mkdir(mdrop);
}
#endif


void crlf_mdir_init() {
	crlf_mdir_exists = 1;
	msgnr = 0;

	/* close an old fd if we have it */

	crlf_mdir_clean();

  /* create 100 struct messages, and messages.md_specific 
   * struct message is in maildrop.h
   *
   * md_specific is the mailbox format specific info, in our case
   * struct crlf_mdir_message (in crlfmaildrop.h), it just has a pointer
   * to the filename.
   */
	if (md_alloc(sizeof(struct crlf_mdir_message)) < 0) {		
		crlf_mdir_release(); /* free() all the structs XXX */
		pop_log(pop_priority, "crlfmaildir: no memory available");
		send_error("no memory available");
		exit(1);
	}
#ifdef CREATEMAILDROP
	if (createmaildrop)
		crlf_mdir_create(); /* XXX */
#endif
	/* if the "new" dir dosn't exist crlf_mdir_exists == 0,
	 * it's one by default */
	crlf_mdir_append("new");
	/* NB we DON'T check the "cur" dir, only "new", so this ISN'T true
	 * maildir. */
#ifndef BULLETINS
	if (crlf_mdir_exists)
		qsort(messages, msgnr, sizeof(struct message),
			crlf_mdir_compare);
#endif
}

void crlf_mdir_release() {
	size_t number;

	if (messages == NULL)
		return;
	/* free the filenames */
	for (number = 0; number < msgnr; number++)
		free(((struct crlf_mdir_message *)
			(messages[number].md_specific))->filename);
	md_free();
}

void crlf_mdir_cleanup(void) {
	close(msgfd);
}

void crlf_mdir_retrieve(unsigned int nr) {
	char msgfile[PATH_MAX];
	ssize_t wrote;
	unsigned char *mmapd_message_buf = NULL;

	if (nr > msgnr) {
		send_error("no such message");
		check_wccount();
		return;
	}
	/* messages start at 1, not 0. */
	if (messages[--nr].deleted) {
		send_error("message %u already marked as deleted", nr + 1);
		check_wccount();
		return;
	}
	if ((strlen(maildrop_name) + strlen(((struct crlf_mdir_message *)
		(messages[nr].md_specific))->filename)
	     + 2) > sizeof(msgfile)) { /* it was already checked in md_append, 
	                            	  but "szczezonego Pan Bog szczeze" */
		pop_log(pop_priority, "crlfmaildir: unexpected error");
		exit(1);
	}
	strcpy(msgfile, maildrop_name);
	strcat(msgfile, "/");
	strcat(msgfile, ((struct crlf_mdir_message *)
		(messages[nr].md_specific))->filename);
	if ((msgfd = open(msgfile, O_RDONLY)) < 0) {
		pop_log(pop_priority, "crlfmaildir: maildir content probably "
			"has been changed");
		pop_error("crlfmaildir: open");
		exit(1);
	}
	if ( (mmapd_message_buf = mmap((caddr_t) 0 , messages[nr].size,
		PROT_READ, MAP_SHARED, msgfd, 0)) == MAP_FAILED) {
		pop_error("crlfmaildir: mmap failed ");
		close(msgfd);
		exit(1);
	}
	messages[nr].cread = 1; /* used to expire read messages */
	send_ok("%u octets", messages[nr].crlfsize);
	if (buf_flush() < 0) {
		close(msgfd);
		exit(1);
	}

/* we don't use skt_write, whole message is written in one call, so buffering
   can only slow it down */
#ifdef SPENCRYPTION
	if (enctrans) {
		wrote = write(encpipe[1], mmapd_message_buf,
			messages[nr].size);
	} else {
		if (encenabled)
			wrote = enc_write(mmapd_message_buf,
				messages[nr].size);
		else
			wrote = write(1, mmapd_message_buf,
				messages[nr].size);
	}
#else
	wrote = write(1, mmapd_message_buf, messages[nr].size);
#endif
	if (wrote < 0) {
		pop_error("crlfmaildir: write failed");
		close(msgfd);
		exit(1);
	}
	if (munmap(mmapd_message_buf, messages[nr].size) != 0) {
		pop_error("crlfmaildir: munmap failed");
		close(msgfd);
		exit(1);
	}
	close(msgfd);
	md_end_reply(&crlf_mdir_cleanup);	
}

void crlf_mdir_update() {
	int tmp;
	char filename[PATH_MAX], *tmp2;

	if (crlf_mdir_exists == 0)
		return;
	if ((strlen(maildrop_name) + 3) > sizeof(filename)) {
			/* no overflow, already checked in crlf_mdir_append but
			   "szczezonego Pan Bog szczeze" */
		crlf_mdir_release();
		pop_log(pop_priority, "crlfmaildir: unexpected error - send "
			"info to author of program");
		send_error("fatal error");
		exit(1);
	}
	strcpy(filename, maildrop_name);
	strcat(filename, "/");
	tmp2 = filename + strlen(filename);
	for (tmp = 0; tmp < msgnr; tmp++) {
		*tmp2 = 0;
		if ((strlen(filename) + strlen(((struct crlf_mdir_message *)
			(messages[tmp].md_specific))->filename) + 1) >
			sizeof(filename)) {
			/* no overflow, already checked in crlf_mdir_append but
			   "szczezonego Pan Bog szczeze" */
			crlf_mdir_release();
			pop_log(pop_priority, "crlfmaildir: unexpected error - "
				"send info to author of program");
			send_error("fatal error");
			exit(1);
		}
		strcpy(tmp2, ((struct crlf_mdir_message *)
			(messages[tmp].md_specific))->filename);
		if (messages[tmp].deleted)
			unlink(filename);
	}
	*tmp2 = 0;
}

void crlf_mdir_top(unsigned int nr, unsigned int lcount) {
	char msgfile[PATH_MAX];
	unsigned char *mmapd_message_buf = NULL;
	ssize_t end_of_lines, wrote;

	if (nr > msgnr) {
		send_error("no such message");
		check_wccount();
		return;
	}
	if (messages[--nr].deleted) {
		send_error("message %u is marked as deleted", nr + 1);
		check_wccount();
		return;
	}
	if ((strlen(maildrop_name) + strlen(((struct crlf_mdir_message *)
		(messages[nr].md_specific))->filename) + 2) >
		sizeof(msgfile)) { /* it was already checked in md_append, 
	                              but "szczezonego Pan Bog szczeze" */
		pop_log(pop_priority, "crlfmaildir: unexpected error");
		exit(1);
	}
	strcpy(msgfile, maildrop_name);
	strcat(msgfile, "/");
	strcat(msgfile, ((struct crlf_mdir_message *)
		(messages[nr].md_specific))->filename);

	if ((msgfd = open(msgfile, O_RDONLY)) < 0) {
		pop_log(pop_priority, "crlfmaildir: maildir content probably "
			"has been changed");
		pop_error("crlfmaildir: open");
		exit(1);
	}
	/*
	 * mmap the file, skip header,
	 * walk through it counting \n
	 * when we get to lcount \n's
	 * write the whole thing out.
	 */

	if ((mmapd_message_buf = mmap((caddr_t) 0 , messages[nr].size,
		PROT_READ, MAP_SHARED, msgfd, 0)) == MAP_FAILED) {
		pop_error("crlfmaildir: mmap failed ");
		close(msgfd);
		exit(1);
	}
	send_ok("");
	if (buf_flush() < 0) {
		close(msgfd);
		exit(1);
	}
	/* skip header */
	wrote = 0;
	if ((lcount < messages[nr].size) && (messages[nr].size > 2))
		/* fetchmail uses TOP <msg> 99999999, which is bigger
		   than most messages */
		for (end_of_lines = 0; end_of_lines < (messages[nr].size - 2);
			end_of_lines++)
			if ((mmapd_message_buf[end_of_lines] == '\n') &&
				(mmapd_message_buf[end_of_lines + 1] == '\r') &&
				(mmapd_message_buf[end_of_lines + 2] == '\n')) {
				end_of_lines += 3;
				wrote = 1;
				break;
			}
	if (!wrote)
		end_of_lines = messages[nr].size;
	if (lcount >= (messages[nr].size - end_of_lines))
		end_of_lines = messages[nr].size;
	else
		for (; (lcount > 0) && (end_of_lines < messages[nr].size);
			end_of_lines++) {
			if (mmapd_message_buf[end_of_lines] == '\n')
				lcount--;
		}
#ifdef SPENCRYPTION
	if (enctrans) {
		wrote = write(encpipe[1], mmapd_message_buf,
			end_of_lines);
	} else {
		if (encenabled)
			wrote = enc_write(mmapd_message_buf,
				end_of_lines);
		else
			wrote = write(1, mmapd_message_buf,
				end_of_lines);
	}
#else
	wrote = write(1, mmapd_message_buf, end_of_lines);
#endif
	if (wrote < 0) {
		pop_error("crlfmaildir: write failed ");
		close(msgfd);
		exit(1);
	}
	if (munmap(mmapd_message_buf, messages[nr].size) != 0) {
		pop_error("crlfmaildir: munmap failed ");
		exit(1);
	}
	md_end_reply(&crlf_mdir_cleanup);
	close(msgfd);
}

/* 
 * uses the filename.
 */
char *crlf_mdir_md5_uidl_message(unsigned int number, char *result) {
	size_t length;
	char *tmp2, *tmp3;
	struct md5_ctx context;
	
	md5_init_ctx(&context);
	tmp2 = ((struct crlf_mdir_message *)
		(messages[number].md_specific))->filename;
	if ((tmp3 = strrchr(tmp2,'/')) == NULL)
		tmp3 = tmp2;
	else
		tmp3 = tmp3 + 1;
	length = strlen(tmp3);
	if ((tmp2 = strchr(tmp3, ':')) != NULL)
		length -= strlen(tmp2);
	md5_process_bytes(tmp3, length, &context);
	md5_finish_ctx(&context, result);
	return result;
}
