static const char rcsid[] = "$Id: into_crlfmaildir.c,v 1.6 2000/09/03 13:28:47 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 2000  Jerzy Balamut <jurekb@dione.ids.pl>
 *  Copyright (C) 2000  Jasper Wallace <jasper@ivision.co.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This program is ment to be popen'd by the MTA delivering mail to local
 * users.
 *
 * it processes the mail and shoves it into the mail box specified as the only
 * argument.
 *
 * It's only usefull if your using crlfmaildir.
 *
 */



#include "includes.h"

#include <errno.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sysexits.h> /* "Exit status codes employed by the mail subsystem" 
                       *  exim (and, i suspect all other MTA's) understands
                       *  EX_TEMPFAIL to mean queue the message rather than
                       *  throwing it away, so we use it to prevent mail
                       *  geting lost.
                       */

void usage (void);
FILE *openmaildir(char *path, char *filename, size_t maxlen, char *hostname);

int main (int argc, char **argv)
{
	int c, nl = 1;
	int crlfsize = 0;
	pid_t tpid = getpid();
	time_t ttime = time(NULL);
	char hostname[256];
	char filename[PATH_MAX];
	char linebuf[128];
	char newname[PATH_MAX];
	char tmppath[PATH_MAX];
	char *path, *result;
	FILE *fh;
	struct passwd *pwentry;

	if (argc != 2)
		usage();
	path = argv[1];
	if (path[0] != '/') {
		if ((pwentry = getpwuid(getuid())) == NULL) {
			perror("unable to get user password entry: getpwuid");
			exit(EX_TEMPFAIL);
		}
		snprintf(tmppath, sizeof(tmppath), "%s/%s", pwentry->pw_dir,
			path);
		path = tmppath;
	}
	if (gethostname(hostname, sizeof(hostname)) < 0) {
		perror("unable to get hostname");
		exit(EX_TEMPFAIL); /* there is also EX_NOHOST for exactly
				      this error */
	}
	
	fh = openmaildir(path, filename, sizeof(filename), hostname);
	if ((result = fgets(linebuf, sizeof(linebuf), stdin)) == NULL) {
		unlink(filename);
		perror("empty message");
		exit(EX_DATAERR);
	}
	/* discard "From " envelope - Postfix's mailbox_command in main.conf */
	if ((strlen(linebuf) > 5) && ((strncmp(linebuf, "From ", 5)) == 0)) {
		while ((linebuf[strlen(linebuf) - 1] != '\n') &&
			(result != NULL))
			result = fgets(linebuf, sizeof(linebuf), stdin);
		if (fgets(linebuf, sizeof(linebuf), stdin) == NULL) {
			unlink(filename);
			perror("corrupted message");
			exit(EX_DATAERR);
		}
	}
	do {
		c = strlen(linebuf); /* c > 0 <==> !EOF */
		if (nl && (linebuf[0] == '.'))
			putc('.', fh);
		if (linebuf[c - 1] == '\n') {
			nl = 1;
			linebuf[--c] = 0;
		} else
			nl = 0;
		fputs(linebuf, fh);
		crlfsize += c;
		if (nl) {
			fputs("\r\n", fh);
			crlfsize += 2;
		}
	} while (fgets(linebuf, sizeof(linebuf), stdin) != NULL);
	fclose(fh);
	if (!nl) {
		unlink(filename);
		perror("corrupted message");
		exit(EX_DATAERR);
	}
	do {
		pid_t nnomsg = 0;
		int ret;
		if ( snprintf(newname, sizeof(newname),
			"%s/new/%lu.%lu.%.256s:%lu", path,
			(unsigned long int)ttime, (unsigned long int)tpid,
			hostname, (unsigned long int)crlfsize)
			> sizeof(newname) ) {
			unlink(filename);
			fprintf(stderr, "into_crlfmaildir: message file name "
				"is too long\n");
			exit(EX_TEMPFAIL);
		}
		ret = link(filename, newname);
		if (( ret < 0) && (errno == EEXIST)) {
			tpid++;
			nnomsg++;
			if (nnomsg > 500) {
				unlink(filename);
				fprintf(stderr, "into_crlfmaildir: can't find "
					"unused message file name\n");
				exit(EX_TEMPFAIL);
			}
			continue; /* try next message file name */
		}
		if (ret < 0) {
			unlink(filename);
			fprintf(stderr, "maildir: can't link message file: "
				"%.1024s", filename);
			exit(EX_TEMPFAIL);
		}
		break;
	} while(1);
	unlink(filename);
	exit(0);
}

FILE *openmaildir(char *path, char *filename, size_t maxlen, char *hostname)
{
	FILE *fh;
	struct stat dstat;
	int x, len, fd;
	pid_t nnomsg = 0;
	pid_t tpid = getpid();
	time_t ttime = time(NULL);
	char t, tmp[PATH_MAX];
	
	fh = stdout;

	umask(0);

	if (stat(path, &dstat) < 0)
	{
		if (errno == ENOENT) {
			/* the dir. dosn't exist, create it. */
			len = strlen(path);
			/* first check for and remove trailing / */
			if (path[len - 1] == '/')
			{
				path[len - 1] = '\0';
				len--;
			}
			if (len < 2)
			{
				fprintf(stderr, "path too short: %s\n", path);
				exit(EX_TEMPFAIL);
			}
			for (x = 0 ; x < len ; x++ )
			{
				if (path[x] == '/')
				{
					t = path[x+1];
					path[x+1] = '\0';
					if (stat(path, &dstat) < 0)
					{
		/* this component dosn't exist, so mkdir it.
		 *
		 * has to be g+rwx so that if a mailbox hashing scheme is
		 * being used other mailboxes can be written.
		 */
						if (mkdir(path, 0770) < 0)
						{
							fprintf(stderr,
								"unable to "
								"mkdir %s : "
								"%s\n", path,
								strerror(errno));
							exit(EX_TEMPFAIL);
						}
					}
					path[x+1] = t;
				}
			}
			/* the last component is the actual mailbox */
			if (stat(path, &dstat) < 0)
			{
			/* only u+rwx - we don't want people in the same group
			 * reading each others mail.
			 */
				if (mkdir(path, 0700) < 0)
				{
					fprintf(stderr, "unable to mkdir %s : "
						"%s\n", path, strerror(errno));
					exit(EX_TEMPFAIL);
				}
			}
		} else {
			fprintf(stderr, "unable to stat %s : %s\n", path,
				strerror(errno));
			exit(EX_TEMPFAIL);
		}
	}
	
	snprintf(tmp, sizeof(tmp), "%s/new", path);
	if (stat(tmp, &dstat) < 0)
	{
		if (mkdir(tmp, 0700) < 0)
		{
			fprintf(stderr, "unable to mkdir %s : %s\n", tmp,
				strerror(errno));
			exit(EX_TEMPFAIL);
		}
	}

	snprintf(tmp, sizeof(tmp), "%s/tmp", path);
	if (stat(tmp, &dstat) < 0)
	{
		if (mkdir(tmp, 0700) < 0)
		{
			fprintf(stderr, "unable to mkdir %s : %s\n", tmp,
				strerror(errno));
			exit(EX_TEMPFAIL);
		}
	}
	do {
		snprintf(filename, maxlen, "%s/tmp/%lu.%lu.%.256s", path,
			(unsigned long int)ttime, (unsigned long int)tpid,
			hostname);
	/* get the perms. right, we could just use chmod i suppose. */
		if (((fd = open(filename, O_WRONLY | O_CREAT | O_EXCL, 0600))
			< 0) && (errno != EEXIST)) {
			fprintf(stderr, "unable to open %s : %s\n", tmp,
				strerror(errno));
			exit(EX_TEMPFAIL);
		}
		if (fd < 0) {
			nnomsg++;
			tpid++;
			if (nnomsg > 500) {
				fprintf(stderr, "into_crlfmaildir: can't find "
					"unused message file name\n");
				exit(EX_TEMPFAIL);
			}
			continue;
		}
		break;
	} while(1);
	if ( (fh = fdopen(fd, "w")) == NULL)
	{
		fprintf(stderr, "unable to fdopen %s : %s\n", tmp,
			strerror(errno));
		exit(EX_TEMPFAIL);
	}
	return(fh);
}

void usage (void)
{
	fprintf(stderr,"Usage: into_crflmaildir /path/to/mailbox\n");
	exit(EX_TEMPFAIL);
}

