static const char rcsid[] = "$Id: spsasl.c,v 1.11 2000/09/13 21:03:17 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include "spsasl.h"
#include "log.h"
#include "response.h"
#include "utils.h"
#include "sktbuff.h"
#include "options.h"
#ifdef SPIPV6
# include "spipv6.h"
#endif
#include <sasl.h>
#include <saslutil.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>

extern void child_corrupted(void);

extern char username[];
extern int tunnel[];
extern int first_response;
extern int encenabled;
extern int enctrans;
extern void check_wccount(void);
extern int auth_finished;
#ifdef MAPPING
extern char mapusername[];
extern int domapping;
extern void do_mapusername(void);
#endif
#ifdef USERPASSWD
extern int allowuserpasswd;
extern int requserpasswd;
#endif
#ifdef SPTLS
extern int requiretls;
#endif
extern int gargc;
extern char **gargv;
extern char sasl_secret[];
extern int sasl_secret_len;

int authenticate_sasl(char *mech, char *clientresponse);

char sasl_mechanism[256];
int sasl_mechanism_ln;
int sasl_mechanism_count;
sasl_conn_t *sasl_connection;
int adone = 0, adispose = 0;
char respbuff[2048];
size_t respbuff_count = 0;
int sasl_tunnel[2] = {-1, -1};
int sasl_auth_done = 0;

int min_ssf = 0, max_ssf = 10240; /* should enough */
int allowsasl = 1;
char saslrealm[PATH_MAX];

int spsasl_getsecret(void *context, const char *mechanism, 
	const char *auth_identity, const char *realm, sasl_secret_t **secret) {
	int tmp = 1;
	char s_secret[SASLSECRETLN];

	if (sasl_auth_done) /* shouldn't happen */
		return SASL_FAIL;
	if (strlen(auth_identity) > MAXARGLN) {
		pop_log(pop_priority, "spsasl: username is too long");
		send_error("fatal error");
		exit(1);
	}
	strncpy(username, auth_identity, MAXARGLN + 1);
#ifdef MAPPING
	do_mapusername();
	write_general_auth(AUTH_SASL, username, "x", mapusername);
#else
	write_general_auth(AUTH_SASL, username, "x");
#endif
	/* After write_general_auth, we can't use send_ok, send_error
	   and we can't exit, cause there may be need to run encryption
	   server. */
	
	/* we know username */
	if (write_loop(tunnel[1], &tmp, sizeof(tmp)) < 0)
		write_auth_error(); /* general auth ends here */
	if (write_string(tunnel[1], (char *)mechanism) < 0)
		return SASL_FAIL;
	/* read secret from sasl_tunnel */
	if (read_loop(sasl_tunnel[0], &tmp, sizeof(tmp)) != sizeof(tmp))
		return SASL_FAIL;
	if ((tmp < 0) || (tmp > sizeof(s_secret)))
		return SASL_BUFOVER;
	if (read_loop(sasl_tunnel[0], s_secret, tmp) != tmp) {
		memset(s_secret, 0, sizeof(s_secret));
		return SASL_FAIL;
	}
	sasl_close_pipe();
	sasl_auth_done = 1;
	if (tmp == 0)
		return SASL_NOUSER; /* no user or no user/mechanism pair */
	if ((*secret = malloc(sizeof(sasl_secret_t) + tmp)) == NULL) {
		memset(s_secret, 0, sizeof(s_secret));
		return SASL_NOMEM;
	}
	(*secret)->len = tmp;
	memcpy((*secret)->data, s_secret, tmp);
	memset(s_secret, 0, sizeof(s_secret));
	return SASL_OK;
}

void do_auth(char *arg) {
	int tmp, tmp4;
	char *tmp2 = sasl_mechanism, *tmp3;
	char mechname[128];
	char *clientresponse = NULL;
	
	if (!allowsasl) {
		send_error("SASL authentication not allowed");
		check_wccount();
		return;
	}
#ifdef SPTLS
	if (requiretls && ((encenabled & 2) == 0)) {
		send_error("you must establish TLS connection first");
		check_wccount();
		return;
	}
#endif
	if ((tmp3 = strchr(arg, ' ')) != NULL) {
		*tmp3 = 0;
		clientresponse = tmp3 + 1;
	}
	for (tmp = 1; tmp <= sasl_mechanism_count; tmp++) {
		
		if (tmp < sasl_mechanism_count)
			*(tmp3 = strchr(tmp2, ' ')) = 0;
		else
			tmp3 = NULL;
		tmp4 = strcasecmp(tmp2, arg);
		if (tmp4 == 0) {
			mechname[0] = 0;
			strncat(mechname, tmp2, sizeof(mechname) - 1);
			if (strlen(tmp2) >= sizeof(mechname)) {
				/* shouldn't happen */
				send_error("name too long");
				check_wccount();
				return;
			}
		}
	
		if (tmp3 != NULL)
			*tmp3 = ' ';
		if (tmp4 == 0)
			break;
		tmp2 = tmp3 + 1;
	}
	if (tmp4 != 0) {
		send_error("unknown mechanism");
		check_wccount();
		return;
	}
#ifdef USERPASSWD
	allowuserpasswd = requserpasswd = 0; /* SASL is used, not USER/PASS */
#endif
	tmp4 = authenticate_sasl(mechname, clientresponse);
	sasl_close_pipe();
	auth_finished = 1;
	if (!sasl_auth_done) { /* failed or EXTERNAL authentication */
		if (tmp4 == 0) {
#ifdef MAPPING
			write_general_auth(AUTH_SASL, "x", "x", "");
#else
			write_general_auth(AUTH_SASL, "x", "x");
#endif
			/* username is not known */
			tmp = 0;
			if (write_loop(tunnel[1], &tmp, sizeof(tmp)) < 0)
				write_auth_error();
			return;
		}
		/* authenticated through EXTERNAL (or maybe other) mechanism */
#ifdef MAPPING
		domapping = 0; /* no mapping in EXTERNAL mechanism */
		write_general_auth(AUTH_SASL, username, "x", "");
#else
		write_general_auth(AUTH_SASL, username, "x");
#endif
		/* username known, user authenticated, don't send secret */
		tmp = 2;
		if (encenabled)
			tmp |= 4;
		/* EXTERNAL mechanism doesn't need it, but KERBEROSV4 
		   probably does */
		if (write_loop(tunnel[1], &tmp, sizeof(tmp)) < 0)
			write_auth_error();
		return;

	}
	if (encenabled)
		tmp4 |= 4; /* SASL encenabled couldn't be passed
			in write_general_auth, so we pass it here */
	if (write_loop(tunnel[1], &tmp4, sizeof(tmp4)) < 0)
		return; /* can't end after write_general_auth */
}

void init_sasl(void) {
	int result;
	
	if ((result = sasl_server_init(NULL, SERVICE_NAME)) != SASL_OK) {
		pop_log(pop_priority, "spsasl: sasl_server_init: %.512s",
			sasl_errstring(result, NULL, NULL));
		exit(1);
	}
	adone = 1;
}

void setprop_sasl(void *arg, int num) {
	int result;

	if ((result = sasl_setprop(sasl_connection, num, arg)) != SASL_OK) {
		pop_log(pop_priority, "spsasl: sasl_setprop: %.512s",
			sasl_errstring(result, NULL, NULL));
		exit(1);
	}
}

void set_secprops(char *authid, int ssf) {
	sasl_security_properties_t secprops;
	sasl_external_properties_t external;
	int result;
	char *tmp = NULL;

	if (sasl_connection == NULL)
		return;
	secprops.maxbufsize = 2048;
	secprops.min_ssf = min_ssf;
	secprops.max_ssf = max_ssf;
	secprops.security_flags = SASL_SEC_NOPLAINTEXT | SASL_SEC_NOANONYMOUS;
	secprops.property_names = NULL;
	secprops.property_values = NULL;
	setprop_sasl(&secprops, SASL_SEC_PROPS);
	external.auth_id = authid;
	external.ssf = ssf;
	setprop_sasl(&external, SASL_SSF_EXTERNAL);
	if ((result = sasl_listmech(sasl_connection, NULL, NULL, " ", NULL,
				    &tmp, &sasl_mechanism_ln,
				    &sasl_mechanism_count)) != SASL_OK) {
		pop_log(pop_priority, "spsasl: sasl_listmech: %.512s",
			sasl_errstring(result, NULL, NULL));
		exit(1);
	}
	if (sasl_mechanism_ln >= sizeof(sasl_mechanism)) {
		if (tmp != NULL)
			free(tmp);
		pop_log(pop_priority, "spsasl: list of SASL mechanisms is "
			"too long");
		exit(1);
	}
	memcpy(sasl_mechanism, tmp, sasl_mechanism_ln);
	if (tmp != NULL)
		free(tmp);
	sasl_mechanism[sasl_mechanism_ln] = 0;
}

void setprops_sasl(void) {
#ifdef SPIPV6
	union sp_sockaddr sasladdr;
	socklen_t len = sizeof(sasladdr);
	
	if (getpeername(0, (struct sockaddr *) &sasladdr.saddr_in6, &len) < 0) {
		pop_error("spsasl: getpeername");
		exit(1);
	}
	if (sasladdr.saddr_in6.sin6_family == AF_INET)
		setprop_sasl(&sasladdr.saddr_in, SASL_IP_REMOTE);
	else
		pop_log(pop_priority, "spsasl: warning: peer address isn't "
			"an IPv4 address");
	len = sizeof(sasladdr);
	if (getsockname(0, (struct sockaddr *) &sasladdr.saddr_in6, &len) < 0) {
		pop_error("spsasl: getsockname");
		exit(1);
	}
	if (sasladdr.saddr_in6.sin6_family == AF_INET)
		setprop_sasl(&sasladdr.saddr_in, SASL_IP_LOCAL);
	else
		pop_log(pop_priority, "spsasl: warning: my address isn't "
			"an IPv4 address");
#else /* SPIPV6 */
	struct sockaddr_in sasladdr;
	socklen_t len = sizeof(sasladdr);

	if (getpeername(0, (struct sockaddr *) &sasladdr, &len) < 0) {
		pop_error("spsasl: getpeername");
		exit(1);
	}
	setprop_sasl(&sasladdr, SASL_IP_REMOTE);
	len = sizeof(sasladdr);
	if (getsockname(0, (struct sockaddr *) &sasladdr, &len) < 0) {
		pop_error("spsasl: getsockname");
		exit(1);
	}
	setprop_sasl(&sasladdr, SASL_IP_LOCAL);
#endif
	set_secprops(NULL, 0);
}


void new_connection_sasl(void) {
	int result;
	static sasl_callback_t callbacks[] = {
		{SASL_CB_SERVER_GETSECRET, spsasl_getsecret, NULL},
		{SASL_CB_LIST_END, NULL, NULL}
	};

	if ((result = sasl_server_new(SERVICE_NAME, NULL, saslrealm[0] ? 
		saslrealm : NULL , callbacks, SASL_SECURITY_LAYER,
		&sasl_connection)) != SASL_OK) {
		pop_log(pop_priority, "spsasl: sasl_server_new: %.512s",
			sasl_errstring(result, NULL, NULL));
		exit(1);
	}
	adispose = 1;	
}


void dispose_sasl(void) {
	if (adispose)
		sasl_dispose(&sasl_connection);
	adispose = 0;
}

void done_sasl(void) {
	if (adone)
		sasl_done();
	adone = 0;
}

int read_response(char *buf, unsigned int *length) { /* buf should be 2048-byte
	buffer */
	char abuff[2048]; /* = sizeof(abuff in authenticate_sasl()) - 
			     sasl_decode64(...) doesn't overflow buffer */
	ssize_t result;
	
	if ((result = skt_fgets(abuff, sizeof(abuff) - 1)) < 0) {
		pop_log(pop_priority, "spsasl: can't read from socket");
		pop_error("spsasl: read");
		return -1;
	}
	abuff[result] = 0;
	if ((strcmp(abuff, "*\r\n") == 0) || (result < 2))
		return -1;
	if ((abuff[result - 2] != '\r') || (abuff[result - 1] != '\n'))
		return -1;
	if (result > 2) {
		if (sasl_decode64(abuff, result - 2, buf, length) != SASL_OK)
			return -1;
	} else {
		*length = 0;
	}
	return 0;
}

int authenticate_sasl(char *mech, char *clientresponse) {
	char abuff[2048];
	char *serverout = NULL;
	const char *errstr;
	unsigned int decodedlength = 0, serveroutlen, outlen;
	int result;
	sasl_ssf_t *mechssf;
	
	if (clientresponse != NULL) {
		if (strlen(clientresponse) >= sizeof(abuff)) {
			/* it shouldn't happen */
			pop_log(pop_priority, "spsasl: weird error - "
				"contact author of program");
			send_error("fatal error");
			exit(1);
		}
		if (strlen(clientresponse) > 0)
			if (sasl_decode64(clientresponse,
				strlen(clientresponse), abuff, 
				&decodedlength) != SASL_OK) {
				pop_log(pop_priority, "spsasl: can't decode "
					"client response");
				send_error("fatal error");
				exit(1);
		}
	}
#if 0 /* workaround of bug in cyrus-sasl */
	if ((result = sasl_server_start(sasl_connection, mech, NULL, 0,
		&serverout, &serveroutlen, &errstr)) != SASL_CONTINUE) {
		if (serverout != NULL)
			free(serverout);
		pop_log(pop_priority, "spsasl: workaround failed #1");
		send_error("fatal error");
		exit(1);
	}
	if ((serveroutlen > 0) && (decodedlength > 0)) {
		if (serverout != NULL)
			free(serverout);
		pop_log(pop_priority, "spsasl: workaround failed #2");
		send_error("fatal error");
		exit(1);
	}
	if (serveroutlen == 0) {
		if (serverout != NULL)
			free(serverout);
		result = sasl_server_step(sasl_connection, abuff, decodedlength,
			&serverout, &serveroutlen, &errstr);
	}
#else
	result = sasl_server_start(sasl_connection, mech, abuff, decodedlength,
		&serverout, &serveroutlen, &errstr);
#endif /* workaround */
	while (result == SASL_CONTINUE) {
		strcpy(abuff, "+ ");
		if (serveroutlen > 0)
			if (sasl_encode64(serverout, serveroutlen, abuff + 2,
				sizeof(abuff) - 5, &outlen) != SASL_OK) {
				pop_log(pop_priority, "spsasl: weird error - "
					"contact author of program");
				if (sasl_auth_done)
					return 0;
				else {
					send_error("fatal error");
					exit(1);
				}
			}
		if (serverout != NULL)
			free(serverout);
		abuff[outlen + 2] = 0;
		strncat(abuff, "\r\n", sizeof(abuff) - 1);
		if (skt_write(abuff, strlen(abuff)) < 0) {
			pop_log(pop_priority, "spsasl: can't write to socket");
			pop_error("spsasl: write");
			if (sasl_auth_done)
				return 0;
			else
				exit(1);
		}
		if (buf_flush() < 0)
			exit(1);
		if (read_response(abuff, &decodedlength) < 0) {
			if (sasl_auth_done)
				return 0;
			else
				exit(1);
		}
		result = sasl_server_step(sasl_connection, abuff,
			decodedlength, &serverout, &serveroutlen, &errstr);
	}
	if (serverout != NULL)
		free(serverout);
	if (result != SASL_OK) {
		if ((result != SASL_BADAUTH) && (result != SASL_NOUSER)) {
			if (errstr == NULL)
				errstr = sasl_errstring(result, NULL, NULL);
			if (errstr == NULL)
				errstr = "unknown error";
			pop_log(pop_priority, "spsasl: %.512s", errstr);
			if (sasl_auth_done)
				return 0;
			else
				exit(1);
			
		} else
			return 0;
	}
	if ((result = sasl_getprop(sasl_connection, SASL_USERNAME,
		(void **) &serverout)) != SASL_OK) {
		pop_log(pop_priority, "spsasl: %.512s",
			sasl_errstring(result, NULL, NULL));
		if (sasl_auth_done)
			return 0;
		else {
			send_error("fatal error");
			exit(1);
		}
	}
	if ((result = sasl_getprop(sasl_connection, SASL_SSF,
		(void **) &mechssf)) != SASL_OK) {
		pop_log(pop_priority, "spsasl: %.512s",
			sasl_errstring(result, NULL, NULL));
		if (sasl_auth_done)
			return 0;
		else {
			send_error("fatal error");
			exit(1);
		}
	}
	if (strlen(serverout) > MAXARGLN) {
		pop_log(pop_priority, "spsasl: username is too long");
		if (sasl_auth_done)
			return 0;
		else {
			send_error("fatal error");
			exit(1);
		}
		
	}
	strncpy(username, serverout, MAXARGLN + 1);
	encenabled |= (((*mechssf) == 0) ? 0 : 1);
	first_response = 1;
	return 1;
}

void sasl_pipe(void) {
	if (pipe(sasl_tunnel) < 0) {
		pop_error("pipe");
		exit(1);
	}
}

void sasl_close_pipe(void) {
	if (sasl_tunnel[0] >= 0)
		close(sasl_tunnel[0]);
	if (sasl_tunnel[1] >= 0)
		close(sasl_tunnel[1]);
	sasl_tunnel[0] = sasl_tunnel[1] = -1;
}

void sasl_client_pipe(void) {
	if (sasl_tunnel[1] >= 0)
		close(sasl_tunnel[1]);
	sasl_tunnel[1] = -1;
}

void sasl_server_pipe(void) {
	if (sasl_tunnel[0] >= 0)
		close(sasl_tunnel[0]);
	sasl_tunnel[0] = -1;
}

int sasl_authenticate(void) {
	char mech[128];
	int len = sasl_secret_len, result;
	char *act = sasl_secret;
	unsigned long int tmp;

	if (read_string(tunnel[0], mech, sizeof(mech)) < 0)
		return -1;
	while (len > 0) {
		if (memchr(act, 0, len) == NULL) {
			memset(sasl_secret, 0, sasl_secret_len);
			pop_log(pop_priority, "spsasl: secret is damaged");
			return -1;
		}
		if ((strlen(act) + 1 + sizeof(unsigned long int)) > len) {
			memset(sasl_secret, 0, sasl_secret_len);
			pop_log(pop_priority, "spsasl: secret is damaged");
			return -1;
		}
		result = strcmp(act, mech);
		len -= (strlen(act) + 1 + sizeof(unsigned long int));
		act += (strlen(act) + 1);
		tmp = *((unsigned long int *) act);
		act += sizeof(unsigned long int);
		if (tmp > len) {
			memset(sasl_secret, 0, sasl_secret_len);
			pop_log(pop_priority, "spsasl: secret is damaged");
			return -1;
		}
		len -= tmp;
		if (result == 0)
			break;
		act += tmp;
	}
	len = ((result != 0) ? 0 : tmp);
	if (write_loop(sasl_tunnel[1], &len, sizeof(len)) < 0) {
		memset(sasl_secret, 0, sasl_secret_len);
		return -1;
	}
	if (len > 0)
		if (write_loop(sasl_tunnel[1], act, tmp) < 0) {
			memset(sasl_secret, 0, sasl_secret_len);
			return -1;
		}
	sasl_close_pipe();
	memset(sasl_secret, 0, sasl_secret_len);
	if (result < 0)
		return -1;
	if (read_loop(tunnel[0], &result, sizeof(result)) != sizeof(result))
		return -1;
	if (((result & ~4) != 0) && ((result & ~4) != 1))
		return -1;
	if (result & 4)
		encenabled = enctrans = 1;
	else
		close_encpipes();
	return (((result & 1) == 0) ? -1: 0);
}
